// SDIse.java
// Speciation versus Duplication Inference
//
// Copyright (C) 1999-2001 Washington University School of Medicine
// and Howard Hughes Medical Institute
// All rights reserved
//
// Created: 2000
// Author: Christian M. Zmasek
// zmasek@genetics.wustl.edu
// http://www.genetics.wustl.edu/eddy/people/zmasek/


// Requires JDK 1.2 or greater.


package forester.tools;


import forester.tree.*;

import java.util.HashMap;      // Requires JDK 1.2 or greater.



/**

Implements our algorithm for speciation - duplication inference (SDI).

<p>
Reference:
</p>
<ul>
<li>
Zmasek, C.M. and Eddy, S.R. (2001) "A simple algorithm to infer gene duplication 
and speciation events on a gene tree". Bioinformatics, in press.
</ul>

<p>
The initialization is accomplished by:
</p>
<ul>
<li>
method "linkExtNodesOfG()" of class SDI: setting the links for
the external nodes of the gene tree
<li>
"preorderReID(int)" from class Tree: numbering of nodes of the species tree
in preorder
<li>
the optional stripping of the species tree is accomplished by
method "stripTree(Tree,Tree)" of class Tree
</ul>

<p>
The recursion part is accomplished by this class' method
"geneTreePostOrderTraversal(Node)".

<p>
Requires JDK 1.2 or greater.

@see SDI#linkExtNodesOfG()
@see Tree#preorderReID(int)
@see SDI#stripTree(Tree,Tree)
@see #geneTreePostOrderTraversal(Node)

@author Christian M. Zmasek

@version 1.101 -- last modified: 06/13/01

*/
public class SDIse extends SDI {
    


    /**

    Constructor which sets the gene tree and the species tree to be
    compared. species_tree is the species tree to which the gene
    tree gene_tree will be compared to - with method "infer(boolean)".
    Both Trees must be completely binary and rooted.
    The actual inference is accomplished with method "infer(boolean)".
    The mapping cost L can then be calculated with method "computeMappingCost()".
    <p>
    (Last modified: 01/11/01)

    @see #infer(boolean)
    @see SDI#computeMappingCost()

    @param gene_tree    reference to a rooted binary gene Tree to which assign
                        duplication vs speciation, must have species names in 
                        the species name fields for all external nodes 
    @param species_tree reference to a rooted binary species Tree which might
                        get stripped in the process, must have species names in 
                        the species name fields for all external nodes 

    */
    public SDIse( Tree gene_tree, Tree species_tree ) throws Exception {

        super( gene_tree, species_tree );

    } // SDIse( Tree, Tree ) 


    
    /**
    
    Infers for each Node of gene_tree (set in constructor) whether it
    represents a speciation or duplication event by calculating 
    and interpreting the mapping function M. The most parsimonious
    sequence of speciation and duplication events is assumed.
    <p>
    (Last modified: 01/11/01)

    @see #SDIse(Tree,Tree)

    @param strip_species_tree set to true to remove from the species tree 
                              species not found in the gene tree prior
                              to analysis (optional)

    @return number of duplications which have been assigned in gene_tree

    */
    public int infer( boolean strip_species_tree ) throws Exception {
        
        duplications = 0;
        
        if ( strip_species_tree ) {
            stripTree( genetree, speciestree );
        }
      
        speciestree.preorderReID( 0 );
        
        linkExtNodesOfG();
       
        geneTreePostOrderTraversal( genetree.getRoot() );
       
        return duplications;

    } // infer( boolean )

    

    /**

    Traverses the subtree of Node g in postorder, calculating the
    mapping function M, and determines which nodes
    represent speciation events and which ones duplication
    events.
    <p>
    Preconditions: Mapping M for external nodes must have been calculated
    and the species tree must be labelled in preorder.
    <p>
    (Last modified: 01/11/01)

    @param g starting node of a gene tree - normally the root

    */
    void geneTreePostOrderTraversal( Node g ) {

        Node a, b;

        if ( !g.isExternal() ) {

            geneTreePostOrderTraversal( g.getChild1() );
            geneTreePostOrderTraversal( g.getChild2() );

            a = g.getChild1().getLink();
            b = g.getChild2().getLink();

            while ( a != b ) {
                if ( a.getID() > b.getID() ) {
                    a = a.getParent();
                }
                else {
                    b = b.getParent();
                }
            }
            g.setLink( a );
            
            // Determines whether dup. or spec.
            g.setDuplicationOrSpecAssigned( true );

            if ( a == g.getChild1().getLink()
            ||   a == g.getChild2().getLink() ) {
                g.setDuplication( true );
                duplications++;
            }
            else {
                g.setDuplication( false );
            }
        }

    } // geneTreePostOrderTraversal( Node )



    /**

    Calculates the mapping for the root of the gene tree
    and its two children.
    <p>
    Is to be used ONLY by method "SDIunrooted.fastInfer(Tree,Tree)".
    <p>
    (Last modfied: 06/12/01)

    @return number of duplications which have been assigned in gene_tree

    */
    int inferR( boolean was_old_root_dup ) throws Exception {
       
        Node    root     = null,
                a        = null,
                b        = null;

        root = genetree.getRoot();

        calculateMforChildrenOfRoot( root.getChild1() );

        calculateMforChildrenOfRoot( root.getChild2() );

        
        a = root.getChild1().getLink();
        b = root.getChild2().getLink();
        while ( a != b ) {
            if ( a.getID() > b.getID() ) {
                a = a.getParent();
            }
            else {
                b = b.getParent();
            }
        }
        root.setLink( a );
        if ( a == root.getChild1().getLink()
        ||   a == root.getChild2().getLink() ) {
            root.setDuplication( true );
            if ( !was_old_root_dup ) {
                duplications++;
            }
        }
        else {
            root.setDuplication( false );
            if ( was_old_root_dup ) {
                duplications--;
            }
            
        }

        return duplications;

    } // inferR( boolean )



    // Helper method for inferR( boolean )
    // (Last modified: 06/12/01)
    private void calculateMforChildrenOfRoot( Node current ) { 
        
        boolean previous = false;

        Node a           = null,
             b           = null;
       
        if ( !current.isExternal() ) {

            a = current.getChild1().getLink();
            b = current.getChild2().getLink();

            while ( a != b ) {
                if ( a.getID() > b.getID() ) {
                    a = a.getParent();
                }
                else {
                    b = b.getParent();
                }
            }
            current.setLink( a );
            previous = current.isDuplication();
            if ( a == current.getChild1().getLink()
            ||   a == current.getChild2().getLink() ) {
                current.setDuplication( true );
                if ( !previous ) {
                    duplications++;
                }
            }
            else {
                current.setDuplication( false );
                if ( previous ) {
                    duplications--;
                }
            }
        }

    } // calculateMforChildrenOfRoot( Node )


    
} // End of class SDIse.
